"""
Visualization module for CORE

Contains visualization functions for:
- Registration results visualization
- Nuclei overlay plots
- Bokeh interactive plots
- Cluster alignment visualization
"""

from core.visualization.visualization import (
    setup_bokeh_notebook,
    visualize_cluster_alignment,
    visualize_overlays,
    visualize_patches,
    create_nuclei_overlay_plot,
    create_registration_comparison_plot,
    visualize_shape_aware_registration,
)

__all__ = [
    "setup_bokeh_notebook",
    "visualize_cluster_alignment",
    "visualize_overlays",
    "visualize_patches",
    "create_nuclei_overlay_plot",
    "create_registration_comparison_plot",
    "visualize_shape_aware_registration",
]
