"""
Evaluation module for CORE

Contains evaluation metrics and functions for:
- Target Registration Error (TRE)
- Normalized Gradient Field (NGF) metric
- Landmark-based evaluation
- Displacement field evaluation
"""

from core.evaluation.evaluation import (
    tre,
    rtre,
    load_landmark_points,
    load_evaluation_landmarks,
    transform_points_homogeneous,
    evaluate_registration_tre,
    apply_displacement_field_to_points,
    evaluate_nonrigid_registration,
    ngf_metric,
)

__all__ = [
    "tre",
    "rtre",
    "load_landmark_points",
    "load_evaluation_landmarks",
    "transform_points_homogeneous",
    "evaluate_registration_tre",
    "apply_displacement_field_to_points",
    "evaluate_nonrigid_registration",
    "ngf_metric",
]
